#include "common.h"

/////////// Mesh ///////////////
//

Mesh::Mesh()
{
	data = NULL;
	faceNb = 0;
}

bool Mesh::Load(const char *filename, int texId)
{
	this->texId = texId;

	FILE *f = fopen(filename, "rt");
	if(!f)
		return false;

	fscanf(f, "%i", &faceNb);
	data = new float[faceNb*3*8];

	for(int i=0; i<faceNb*3*8; i++)
		fscanf(f, "%f", &data[i]);

	fclose(f);

	return true;
}

Mesh::~Mesh()
{
	delete[] data;
}

void Mesh::Render()
{
	glBindTexture(GL_TEXTURE_2D, texId);

	glInterleavedArrays(GL_T2F_N3F_V3F, 0, data);
	glDrawArrays(GL_TRIANGLES, 0, faceNb*3);

	//glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	//glDisableClientState(GL_NORMAL_ARRAY);
	//glDisableClientState(GL_VERTEX_ARRAY);
}

void Mesh::Interpolate(Mesh *m1, Mesh *m2, float alpha)
{
	float invalpha = 1.f - alpha;
	for(int i=0; i<faceNb*3; i++)
	{
		data[i*8+2] = m1->data[i*8+2] * invalpha + m2->data[i*8+2] * alpha;
		data[i*8+3] = m1->data[i*8+3] * invalpha + m2->data[i*8+3] * alpha;
		data[i*8+4] = m1->data[i*8+4] * invalpha + m2->data[i*8+4] * alpha;
		data[i*8+5] = m1->data[i*8+5] * invalpha + m2->data[i*8+5] * alpha;
		data[i*8+6] = m1->data[i*8+6] * invalpha + m2->data[i*8+6] * alpha;
		data[i*8+7] = m1->data[i*8+7] * invalpha + m2->data[i*8+7] * alpha;
	}
}

/////////// AniMesh ///////////////
//

AniMesh::AniMesh()
{
}
AniMesh::~AniMesh()
{
}

bool AniMesh::Load(const char *filename, int texId)
{
	char buffer[256];

	FILE *f = fopen(filename, "rt");
	if(!f)
		return false;

	fscanf(f, "%i", &frameNb);

	for(int i=0; i<frameNb; i++)
	{
		fscanf(f, "%s", &buffer);
		frames[i].Load(buffer, texId);
	}
	renderframe.Load(buffer, texId);

	fclose(f);
	return true;
}

/////////// AniMesh ///////////////
//

AniObject::AniObject(AniMesh *_aniMesh)
{
	aniMesh = _aniMesh;
	animFirst = 0;
	animNb = 1;
	curTime = 0.f;
	animSpeed = 1.f;
}

AniObject::~AniObject()
{
}

void AniObject::SetAnim(int first, int nb, float fps)
{
	curTime = 0.f;
	animFirst = first;
	animNb = nb;
	animSpeed = fps;
}

void AniObject::Update(float ticks)
{
	curTime +=  ticks;
}

void AniObject::Render()
{
	if(animNb<=1)
		aniMesh->frames[animFirst].Render();
	else
	{
		float frame = curTime * animSpeed;
		int frameOne = int(floorf(frame)) % animNb;
		int frameTwo = (int(frame)+1) % animNb;

		float alpha = fmodf(frame, 1.f);

		aniMesh->renderframe.Interpolate(&aniMesh->frames[frameOne], &aniMesh->frames[frameTwo], alpha);
		glEnable(GL_NORMALIZE);
		aniMesh->renderframe.Render();
		glDisable(GL_NORMALIZE);
	}
}

/////////// Font ///////////////
//

void Font::Load(const char *fontname, int size)
{
	SDL_SysWMinfo sys_info;
	SDL_VERSION(&sys_info.version);
	SDL_GetWMInfo(&sys_info);

	HDC hdc = GetDC(sys_info.window);

	HFONT font, oldfont;
	listBase = glGenLists(96);

	font = CreateFont(-size, 0,0,0, FW_BOLD, 0,0,0, ANSI_CHARSET, OUT_TT_PRECIS, CLIP_DEFAULT_PRECIS, ANTIALIASED_QUALITY, FF_DONTCARE|DEFAULT_PITCH, fontname);
	oldfont = (HFONT) SelectObject(hdc, font);
	wglUseFontBitmaps(hdc, 32, 96, listBase);
	SelectObject(hdc, oldfont);
	DeleteObject(font);
}

void Font::Print(float x, float y, const char *text, ...)
{
	char buffer[1024];

	va_list args;
	va_start(args, text);
	vsprintf(buffer, text, args);
	va_end(args);

	glPushMatrix();
	glRasterPos2f(x,y);

	glPushAttrib(GL_LIST_BIT);
	glListBase(listBase-32);
	glCallLists((GLsizei) strlen(text), GL_UNSIGNED_BYTE, buffer);
	glPopAttrib();

	glPopMatrix();
}
