#include "common.h"

///// GLOBAL SETTINGS /////

//#ifdef _DEBUG
int gWidth = 800;
int gHeight = 600;
//#else
//int gWidth = 1024;
//int gHeight = 768;
//#endif

float gNear = 10.f;
float gFar = 10000.f;
float gFov = 60.f;

#define SCROLL_SPEED 400
#define SCROLL_ACCEL 5
#define ANGLE_SPEED 100
#define ANGLE_ACCEL 5
#define PITCH_SPEED 100
#define PITCH_ACCEL 5
#define ZOOM_SPEED 400
#define ZOOM_ACCEL 5

///// GLOBAL DATA /////

Uint8 *gKeys;

int gGameMode = GMODE_NORMAL;
int gEditorTile = 0;
int gEditorMaxTile = 0;
Camera gEditorCamera;
Font gFont, gBigFont;

///// current game state

Map *gCurMap = NULL;
int gCurLevel = 0;

Camera gCamera;
LightTracer *gLightTracer;
bool gShowHelp = false;

/////////////////////////

int gMouseX;
int gMouseY;
float gMouseXNorm;
float gMouseYNorm;
int gMouseButtons;

const char *gTextureFiles[] = 
{
	"data/wall.jpg",
	"data/wall_cross.jpg",
	"data/red.png",
	"data/green.png",
	"data/blue.png",
	"data/monster_red.jpg",
	"data/monster_green.jpg",
	"data/monster_blue.jpg",
	"data/blockwhite.png",
	"data/hero.jpg",
	"data/crate.jpg",
	NULL
};

int gTextures[MAX_TEXTURES];

struct MeshData
{
	const char *filename;
	int texture;
};


MeshData gMeshData[] = 
{
	{ "data/floor.msh", 0 },
	{ "data/wall.msh", 0 },
	{ "data/hole.msh", 0 },
	{ "data/start.msh", 0 },
	{ "data/mirror_a.msh", 8 },
	{ "data/mirror_b.msh", 8 },
	{ "data/crate.msh", 10 },
	{ "data/prism_a.msh", 8 },
	{ "data/prism_b.msh", 8 },
	{ "data/prism_c.msh", 8 },
	{ "data/prism_d.msh", 8 },
	{ "data/lightsource.msh", 0 },
	{ "data/lightsource.msh", 0 },
	{ "data/lightsource.msh", 0 },
	{ "data/lightsource.msh", 0 },
	{ "data/filter.msh", 2 },
	{ "data/filter.msh", 3 },
	{ "data/filter.msh", 4 },
	{ "data/crate.msh", 8 },
	{ "data/crate.msh", 2 },
	{ "data/crate.msh", 3 },
	{ "data/crate.msh", 4 },
	{ "data/goal.msh", 1 },
	{ "data/start.msh", 0 },
	{ "data/start.msh", 0 },
	{ "data/start.msh", 0 },
	{ NULL, 0}
};
Mesh gMeshes[MAX_MESHES];

TileInfo gTileInfo[] = 
{
	{ "Floor", true, false, true},
	{ "Wall", false, false, false},
	{ "Hole", false, false, true},
	{ "Player", true, false, true},
	{ "Mirror 1/2", false, true, false},
	{ "Mirror 2/2", false, true, false},
	{ "Crate", false, true, false},
	{ "Prism A", false, true, false},
	{ "Prism B", false, true, false},
	{ "Prism C", false, true, false},
	{ "Prism D", false, true, false},
	{ "LightsourceA", false, false, false},
	{ "LightsourceB", false, false, false},
	{ "LightsourceC", false, false, false},
	{ "LightsourceD", false, false, false},
	{ "FilterR", false, true, false},
	{ "FilterG", false, true, false},
	{ "FilterB", false, true, false},
	{ "BlockW", false, true, false},
	{ "BlockR", false, true, false},
	{ "BlockG", false, true, false},
	{ "BlockB", false, true, false},
	{ "Exit", true, false, true},
	{ "Monster R", true, false, true},
	{ "Monster G", true, false, true},
	{ "Monster B", true, false, true},
	{ NULL }
};


MeshData gAniMeshData[] = 
{
	{ "data/monster.ani", 5 },
	{ "data/monster.ani", 6 },
	{ "data/monster.ani", 7 },
	{ "data/player.ani", 9 },
	{ NULL, 0}
};

AniMesh gAniMeshes[MAX_MESHES];


struct LevelData
{
	const char *filename;
};

LevelData gLevelData[] = 
{
	{ "data/map002.map"},
	{ "data/map003.map"},
	{ "data/map004.map"},
	{ "data/map000.map"},
	{ "data/map005.map"},
	{ "data/map001.map"},
	{ "data/map008.map"},
	{ "data/map007.map"},
	{ "data/map006.map"},
	{ NULL }
};


float LightAmbient[] = {40/255.f, 40/255.f, 40/255.f, .5f};
float Light[] = {255/255.f*3, 207/255.f*3, 165/255.f*3, .5f};
float LightPos[] = {900, 100, 800, 1.f};

float Light2[] = {165/255.f*.3f, 207/255.f*.3f, 240/255.f*.3f, .0f};
float LightPos2[] = {-900, 1000, -800, 1.f};

bool LoadTexture(const char *filename, int *id)
{
	ILuint i;
	ilGenImages(1, &i);
	ilBindImage(i);
	if(!ilLoadImage((ILstring) filename))
	{
		ilDeleteImages(1, &i);
		return false;
	}

	*id = ilutGLBindMipmaps();

	ilDeleteImages(1, &i);

	return true;
}

bool LoadTextures()
{
	for(int i=0; gTextureFiles[i]; i++)
		if(!LoadTexture(gTextureFiles[i], &gTextures[i]))
			return false;

	return true;
}

bool LoadMeshes()
{
	int i;
	for(i=0; gMeshData[i].filename; i++)
		if(!gMeshes[i].Load(gMeshData[i].filename , gTextures[gMeshData[i].texture]))
			return false;

	gEditorMaxTile = i;

	return true;
}

bool LoadAniMeshes()
{
	for(int i=0; gAniMeshData[i].filename; i++)
		if(!gAniMeshes[i].Load(gAniMeshData[i].filename , gTextures[gAniMeshData[i].texture]))
			return false;

	return true;
}

void PrepareLevel()
{
	for(Objects::iterator i=gObjects.begin(); i!=gObjects.end(); i++)
		delete *i;
	gObjects.clear();

	for(Monsters::iterator i=gMonsters.begin(); i!=gMonsters.end(); i++)
		delete *i;
	gMonsters.clear();

	gPlayer = new Player(&gAniMeshes[3]);

	gLightTracer->Clear();

	for(Coord c(0,0); c.x<gCurMap->width; c.x++)
		for(c.y=0; c.y<gCurMap->height; c.y++)
	{
		int tile = gCurMap->Get(c);
		switch(tile)
		{
		case TILE_MONSTER_START_R:
		case TILE_MONSTER_START_G:
		case TILE_MONSTER_START_B:
			{
				Monster *monster = new Monster(&gAniMeshes[tile-TILE_MONSTER_START_R]);
				monster->color = (tile-TILE_MONSTER_START_R)+1;
				monster->SetCoord(c);
				gMonsters.push_back(monster);
			}
			break;
		case TILE_PLAYER_START:
			gPlayer->SetCoord(c);
			gMonsters.push_back(gPlayer);
			break;
		case TILE_LIGHTSOURCE_A:
		case TILE_LIGHTSOURCE_B:
		case TILE_LIGHTSOURCE_C:
		case TILE_LIGHTSOURCE_D:
			gLightTracer->AddSource(c, tile-TILE_LIGHTSOURCE_A);
			break;

		default:
			if(gTileInfo[tile].is_object)
			{
				gCurMap->Set(c, TILE_FLOOR);
				Object *obj = new Object(tile);
				obj->SetCoord(c);
				gObjects.push_back(obj);
			}
			break;
		}
	}
	gLightTracer->Update();
}

bool LoadLevel(int i)
{
	delete gCurMap;

	gCurMap = new Map();
	gCurMap->Load(gLevelData[0].filename);

	PrepareLevel();

	return true;
}



Camera::Camera()
{
	angle_speed = zoom_speed = pitch_speed = 0.f;
	//angle = 180;
	//pitch = 45;
	//zoom = 200;
	//pos = Vector(0, 50, 0);
	angle = 0;
	pitch = 70;
	zoom = 800;
	pos = Vector(300, 0, 300);
}

void Camera::Transform()
{
	glLoadIdentity();

	glTranslatef(0,0,-zoom);
	glRotatef(pitch,1,0,0);
	glRotatef(angle,0,1,0);
	glTranslatef(-pos.x,-pos.y,-pos.z);
}

Vector Camera::GetPos()
{
	Vector p(0,0,zoom);
	p = p.RotateX(pitch);
	p = p.RotateY(angle);
	p += pos;
	return p;
}

Vector Camera::GetMouseVector()
{
	GLdouble model[16], proj[16];
	GLint viewport[4];

	glGetDoublev(GL_MODELVIEW_MATRIX, model);
	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	glGetIntegerv(GL_VIEWPORT, viewport);

	GLdouble x,y,z, x2, y2, z2;
	gluUnProject(gMouseX, viewport[3]-gMouseY, 0.f, model, proj, viewport, &x, &y, &z);
	gluUnProject(gMouseX, viewport[3]-gMouseY, 1.f, model, proj, viewport, &x2, &y2, &z2);

	return normal(Vector(float(x2-x),float(y2-y),float(z2-z)));
}


void Camera::Update(float ticks)
{
	if(gGameMode==GMODE_EDITOR)
	{
		if(gKeys[SDLK_UP])
			smooth_change(speed.z, -SCROLL_SPEED, SCROLL_ACCEL, ticks);
		else if(gKeys[SDLK_DOWN])
			smooth_change(speed.z, SCROLL_SPEED, SCROLL_ACCEL, ticks);
		else
			smooth_change(speed.z, 0, SCROLL_ACCEL, ticks);

		if(gKeys[SDLK_LEFT])
			smooth_change(speed.x, -SCROLL_SPEED, SCROLL_ACCEL, ticks);
		else if(gKeys[SDLK_RIGHT])
			smooth_change(speed.x, SCROLL_SPEED, SCROLL_ACCEL, ticks);
		else
			smooth_change(speed.x, 0, SCROLL_ACCEL, ticks);

		if(gKeys[SDLK_a])
			smooth_change(angle_speed, -ANGLE_SPEED, ANGLE_ACCEL, ticks);
		else if(gKeys[SDLK_d])
			smooth_change(angle_speed, ANGLE_SPEED, ANGLE_ACCEL, ticks);
		else
			smooth_change(angle_speed, 0, ANGLE_ACCEL, ticks);

		if(gKeys[SDLK_f])
			smooth_change(pitch_speed, -PITCH_SPEED, PITCH_ACCEL, ticks);
		else if(gKeys[SDLK_r])
			smooth_change(pitch_speed, PITCH_SPEED, PITCH_ACCEL, ticks);
		else
			smooth_change(pitch_speed, 0, PITCH_ACCEL, ticks);

		if(gKeys[SDLK_w])
			smooth_change(zoom_speed, -ZOOM_SPEED, ZOOM_ACCEL, ticks);
		else if(gKeys[SDLK_s])
			smooth_change(zoom_speed, ZOOM_SPEED, ZOOM_ACCEL, ticks);
		else
			smooth_change(zoom_speed, 0, ZOOM_ACCEL, ticks);
	}
	else if(gGameMode==GMODE_NORMAL)
	{
		//if(gKeys[SDLK_w])
		//	smooth_change(pitch_speed, -PITCH_SPEED, PITCH_ACCEL, ticks);
		//else if(gKeys[SDLK_s])
		//	smooth_change(pitch_speed, PITCH_SPEED, PITCH_ACCEL, ticks);
		//else
		//	smooth_change(pitch_speed, 0, PITCH_ACCEL, ticks);

		if(gKeys[SDLK_q])
			smooth_change(zoom_speed, -ZOOM_SPEED, ZOOM_ACCEL, ticks);
		else if(gKeys[SDLK_a])
			smooth_change(zoom_speed, ZOOM_SPEED, ZOOM_ACCEL, ticks);
		else
			smooth_change(zoom_speed, 0, ZOOM_ACCEL, ticks);

		if(zoom<100) zoom=100;
	}

	pos += speed.RotateY(angle) * ticks;
	angle += angle_speed * ticks;
	pitch += pitch_speed * ticks;
	zoom += zoom_speed * ticks;
}


void init_gl()
{
	glClearColor(0,0,0,1);

	glViewport(0, 0, gWidth, gHeight);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(gFov, float(gWidth)/float(gHeight), gNear, gFar);
	glMatrixMode(GL_MODELVIEW);

	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);

	glEnable(GL_LIGHTING);

	glLightfv(GL_LIGHT0, GL_AMBIENT, LightAmbient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, Light);
	glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, .000008f);
	glEnable(GL_LIGHT0);

	glLightfv(GL_LIGHT1, GL_DIFFUSE, Light2);
	glEnable(GL_LIGHT1);

	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	glEnable(GL_TEXTURE_2D);
}

bool game_init()
{
	gTimer.Init();

	init_gl();

	if(!LoadTextures())
		return false;

	if(!LoadMeshes())
		return false;

	if(!LoadAniMeshes())
		return false;

	gFont.Load("Courier New", 20);
	gBigFont.Load("Courier New", 48);

	gLightTracer = new LightTracer;
	LoadLevel(0);

	return true;
}

bool game_loop()
{
	float ticks = gTimer.Tick();

	gKeys = SDL_GetKeyState(NULL);
	gMouseButtons = SDL_GetMouseState(&gMouseX, &gMouseY);
	gMouseXNorm = (2*float(gMouseX) / gWidth) - 1.f;
	gMouseYNorm = (2*float(gMouseY) / gHeight) - 1.f;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


	Vector hotspot;
	if(gGameMode==GMODE_EDITOR)
	{
		gEditorCamera.Update(ticks);
		gEditorCamera.Transform();

		Vector p1 = gEditorCamera.GetPos();
		Vector p2 = gEditorCamera.GetMouseVector();
		hotspot = plane_ray_intersection(p1, p1+p2, Vector(0,1,0), 0);

		LightPos[0] = hotspot.x;
		LightPos[1] = 200;
		LightPos[2] = hotspot.z;
	}
	else
	{
		if(gCurMap->Get(world_to_map(gPlayer->pos))==TILE_EXIT)
		{
			gCurLevel++;
			if(!gLevelData[gCurLevel].filename)
				gCurLevel = 0;

			gCurMap->Load(gLevelData[gCurLevel].filename);
			PrepareLevel();
		}

		for(Monsters::iterator i=gMonsters.begin(); i!=gMonsters.end(); )
		{
			Creature *monster = *(i++); // avoid problems if the object deletes itself
			monster->Update(ticks);
		}

		for(Objects::iterator i=gObjects.begin(); i!=gObjects.end(); )
		{
			Object *obj = *(i++); 
			obj->Update(ticks);
		}

		gCamera.Update(ticks);
		gCamera.pos = gPlayer->pos;
		gCamera.Transform();

		LightPos[0] = gPlayer->pos.x;
		LightPos[1] = 200;
		LightPos[2] = gPlayer->pos.z;

		gLightTracer->Update();
	}

	glLightfv(GL_LIGHT0, GL_POSITION, LightPos);
	glLightfv(GL_LIGHT1, GL_POSITION, LightPos2);


	if(gGameMode==GMODE_EDITOR)
	{
		Coord c = world_to_map(hotspot);

		if(gMouseButtons & SDL_BUTTON(1))
			gCurMap->Set(c, gEditorTile);
		if(gMouseButtons & SDL_BUTTON(3))
			gEditorTile = gCurMap->Get(c);

		gCurMap->Render(c);
	}
	else
	{
		for(Monsters::iterator i=gMonsters.begin(); i!=gMonsters.end(); i++)
			(*i)->Render();

		for(Objects::iterator i=gObjects.begin(); i!=gObjects.end(); i++)
			(*i)->Render();

		gCurMap->Render();

		gLightTracer->Render();
	}


	// Hud

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(0,100.f,0,100.f);
	glMatrixMode(GL_MODELVIEW);

	glLoadIdentity();

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);

	glTranslatef(0,0,-1);

	glColor3f(1,1,1);

	if(gGameMode==GMODE_EDITOR)
	{
		gFont.Print(1,1, "Selected tile: %i %s (change with [])", gEditorTile, gTileInfo[gEditorTile].name );
	}
	else
	{
		gFont.Print(1,1, "%i fps", int(gTimer.GetFps()));
	}

	if(gShowHelp)
	{
		gBigFont.Print(20,85, "Lightstream 48");
		gFont.Print(40,80, "by Codexus");
		gFont.Print(20,75, "Ludum Dare April 2005 Game Competition");
		gFont.Print(5,60, "Goal of the game:");
		gFont.Print(5,55, "The goal is simply to reach the exit of each level");
		gFont.Print(5,52, "as indicated by a red cross. You control the ");
		gFont.Print(5,49, "character with the cursor keys. When you are stuck");
		gFont.Print(5,46, "in a level you can't win anymore press SPACE to ");
		gFont.Print(5,43, "reset it. The levels will introduce the concepts");
		gFont.Print(5,40, "one by one to help you get started. If you want");
		gFont.Print(5,37, "you can skip a level with the + key on the");
		gFont.Print(5,34, "numeric keypas. Good luck!");

		gFont.Print(5,28, "Cursor keys - control the guy");
		gFont.Print(5,25, "SPACE - Reset Level");
		gFont.Print(5,22, "Q/A - Zoom In/Out");
	}
	else
	{
		gFont.Print(45,1, "SPACE=Reset level / F1 for HELP!");
	}


	glEnable(GL_LIGHTING);
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_DEPTH_TEST);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	SDL_GL_SwapBuffers();
	return false;
}

void handle_keydown(int key, int mod)
{
	switch(gGameMode)
	{
	case GMODE_NORMAL:
		if(gShowHelp)
		{
			gShowHelp = false;
			break;
		}

		switch(key)
		{
		case SDLK_F1:
		case SDLK_h:
			gShowHelp = !gShowHelp;
			break;
		case SDLK_HOME:
			gGameMode = GMODE_EDITOR;
			gCurMap->Reload();
			//SDL_ShowCursor(SDL_DISABLE);
			break;
		case SDLK_SPACE:
			gCurMap->Reload();
			PrepareLevel();
			break;

////////////////////////////////////////////////////77
			//CHEAT

		case SDLK_KP_PLUS:
			gCurLevel++;
			if(!gLevelData[gCurLevel].filename)
				gCurLevel = 0;

			gCurMap->Load(gLevelData[gCurLevel].filename);
				PrepareLevel();

			break;

		case SDLK_KP_MINUS:
			gCurLevel--;
			if(gCurLevel<0)
				gCurLevel = 0;

			gCurMap->Load(gLevelData[gCurLevel].filename);
			PrepareLevel();
			break;
//////////////////////////////////////////////////////
		}
		break;
	case GMODE_EDITOR:
		switch(key)
		{
		case SDLK_HOME:
			gGameMode = GMODE_NORMAL;
			PrepareLevel();
			//SDL_ShowCursor(SDL_ENABLE);
			break;
		case SDLK_RIGHTBRACKET:
			gEditorTile = (gEditorTile+1) % gEditorMaxTile;
			break;
		case SDLK_LEFTBRACKET:
			gEditorTile = (gEditorMaxTile + gEditorTile-1) % gEditorMaxTile;
			break;
		case SDLK_KP_PLUS:
			gCurLevel++;
			if(!gLevelData[gCurLevel].filename)
				gCurLevel = 0;

			gCurMap->Load(gLevelData[gCurLevel].filename);
			break;

		case SDLK_KP_MINUS:
			gCurLevel--;
			if(gCurLevel<0)
				gCurLevel = 0;

			gCurMap->Load(gLevelData[gCurLevel].filename);
			break;

		case SDLK_s:
			if((mod&KMOD_CTRL) && (mod&KMOD_SHIFT))
			{
				char buffer[256];

				for(int i=0; ; i++)
				{
					sprintf(buffer, "data/map%03i.map", i);
					FILE *f = fopen(buffer, "rt");
					if(f)
					{
						fclose(f);
						continue;
					}
					else
					{
						gCurMap->Save(buffer);
						break;
					}
				}
			}
			if((mod&KMOD_CTRL) && !(mod&KMOD_SHIFT))
			{
				gCurMap->Save();
				break;
			}

			break;
		}
		break;
	}
}
