#ifndef MAP_H
#define MAP_H

#define TILE_SIZE 100.f

struct Coord
{
	int x, y;

	Coord() {x=y=0;}
	Coord(int _x, int _y) {x = _x; y= _y;}

	Coord operator+(const Coord c)
	{
		return Coord(x+c.x, y+c.y);
	}

	bool operator==(const Coord c) const
	{
		return x==c.x && y==c.y;
	}
};

class Map
{
public:
	char filename[256];
	float tileSize;
	int width, height, layers;
	int *data;

	Map();
	~Map();

	bool Load(const char *filename);
	bool Save(const char *filename=NULL);

	void Reload();

	int Get(Coord c, int l);
	void Set(Coord c, int l, int tile);
	void Clear(int l);

	int Get(Coord c);
	void Set(Coord c, int tile);


	void Render(Coord highlight=Coord(-1,-1));
};


Coord orientation_to_coord(int ori);
Vector orientation_to_vector(int ori);
inline float orientation_to_angle(int ori)
{
	return ori * 90.f;
}

inline Coord world_to_map(Vector v)
{
	return Coord(int(v.x / TILE_SIZE + .5f), int(v.z / TILE_SIZE + .5f));
}

inline Vector map_to_world(Coord c)
{
	return Vector(c.x*TILE_SIZE, 0, c.y*TILE_SIZE);
}

#endif
